#ifndef PARSERCLASS_H
#define PARSERCLASS_H
#include <dirent.h>
#include <string>
#include <sstream>
#include <algorithm>
#include <iterator>
#include <vector>
#include <fstream>
#include <iostream>
#include <map>
#include <bits/stdc++.h>
#include "util.h"
using namespace std;

class Parser{
private:  
    int nOptions;
    int isOption(const char*);
    int inputParser(char**, const string&);
    void cmdLineHelp();
    void replaceString(string&, const string&, const string&);

    vector<vector<string>> helpOpt = {
    {"-n", 
        "<null model>",
        "Path to the file containing null model probabilities of each amino acids (20 amino acids)\n" 
        "Default: data/MML0_estimates_uniprot-all.txt"},
    {"-d", 
        "<compression model>",
        "Path to the file containing conditional probabilities of all possible amino acids (20x20 matrix)\n"
        "Default: data/dam101ConditionalMatrix.txt"},
    {"-p", 
        "<path>",
        "Path to the file containing paths to .fasta files containing input sequences"},
    {"-f", 
        "<input file>",
        "Path to the input file containing multiple sequences in fasta format"},
    {"-c", 
        "[ON|OFF]",
        "Enable checkpoints written to checkpoints/ directory at the end of each temperature step\n"
        "Default: OFF"},
    {"-s", 
        "<starting dictionary>",
        "Path to the checkpoint file\n" 
        "Ex: checkpoints/checkPointData_<temperature>.txt>"},
    {"-u", 
        "[ON|OFF]",
        "Enable usages written to usages/ directory at the end of each temperature step\n" 
        "Default: OFF"},
    {"-o", 
        "[none|minimal|basic|full]",
        "Print outputs to the command line based on one of the given formats\n" 
        "none - output based on the input flag parameters (see below)\n" 
        "minimal - initial outputs and perturbation outputs at every temperature step\n" 
        "basic - initial outputs and perturbation outputs at every 1000 iterations\n" 
        "full - all outputs \n"
        "Default: none"},
    {"-l", 
        "[none|minimal|basic|full]",
        "Write the outputs based on one of the given formats to the log_compass_2.0_<date>.txt file\n" 
        "none - output based on the input flag parameters (see below)\n" 
        "minimal - initial outputs and perturbation outputs at every temperature step\n" 
        "basic - initial outputs and perturbation outputs at every 1000 iterations\n" 
        "full - all outputs \n"
        "Default: basic"},
    {"-T", 
        "<temperature>",
        "Starting temperature of the simulated annealing algorithm\n" 
        "Default: 10000"},
    {"-S", 
        "<stopping temperature>",
        "Stopping temperature of the simulated annealing algorithm\n" 
        "Default: 0.1"},
    {"-P", 
        "<temperature probability>",
        "Probability of temperature reduction of the simulated annealing algorithm\n" 
        "Default: 0.88 (Between [0,1])"},
    {"-I", 
        "<iterations>",
        "Number of perturbations(iterations) at each temperature step\n" 
        "Default: 5000"},
    {"-E", 
        "<tolerance>",
        "Tolerance of temperature difference in acceptance and rejection\n" 
        "Default: 0.000001"},
    {"-epsln", 
        "<epsilon>",
        "Optimisation parameter which bounds the history matrices in alignment algorithm\n" 
        "Default: 10"},
    {"-SUB",	
        "<sequence upper bound>",
		"Upper bound of the length of the sequence in the source collection\n"
		"Default: max"},
	{"-SLB",	
        "<sequence lower bound>",
		"Lower bound of the length of the sequence in the source collection\n"
		"Default: 1"}, 
    {"-pertLegnd", "[ON|OFF]",
        "Type of perturbation being performed and details of the perturbation\n" 
        "Default: OFF"},         
    {"-seqDisect", "[ON|OFF]",
        "Print the dissection of each sequence\n" 
        "Default: OFF"},    
    {"-alignment", "[ON|OFF]",
        "Print the alignment of each sequence\n" 
        "Default: OFF"},      
    {"--help", "",
        "Displays help options and exit.\n"}      
    };

public:
    string nullModel = "data/MML0_estimates_uniprot-all.txt";
    string compModel = "data/dam101ConditionalMatrix.txt";
    string sourcePath = "";
    string sourceFile = "";
    string inptDictionary = "";
    vector<string> cpData;
    double kelvin = 10000;
    double kelvProb = 0.88;
    double stoppingKelv = 0.1; 
    int perturbations = 5000;
    double tolerance = 0.000001;
    int seqUpper = numeric_limits<int>::max();
    int seqLower = 1;
    int epsln = 10;                 //bounding parameter for alignment history matrices
    int iter = 0;
    double checkpointedML = 0;
    
    //Optimization flags
    bool checkpoint = false;
    bool usageFile = false;        // Dissections file (disectFName) - Usages of each concept 
    bool pertLegend = false;      //perturbation legend in add,delete,replace
    bool seqCost = false;   
    bool alignPrint = false;
    int outFlag = 0;
    int logFlag = 2;  

    bool pflag = false;
    bool fflag = false;   

    Parser();
    Parser(int, char **);
};

#endif